@@ -14,8 +14,7 @@
 # You should have received a copy of the GNU General Public License
 # along with this program.  If not, see <http://www.gnu.org/licenses/>.
 
-from __future__ import unicode_literals
-
+import logging
 import sys
 
 from gi.repository import Gio
@@ -24,9 +23,9 @@ from gi.repository import GObject
 from gi.repository import GtkSource
 
 from meld.conf import _
-from meld.misc import fallback_decode
 from meld.settings import bind_settings, meldsettings
-from meld.util.compat import text_type
+
+log = logging.getLogger(__name__)
 
 
 class MeldBuffer(GtkSource.Buffer):
@@ -38,10 +37,9 @@ class MeldBuffer(GtkSource.Buffer):
     )
 
     def __init__(self):
-        GtkSource.Buffer.__init__(self)
+        super().__init__()
         bind_settings(self)
         self.data = MeldBufferData()
-        self.user_action_count = 0
         self.undo_sequence = None
         meldsettings.connect('changed', self.on_setting_changed)
         self.set_style_scheme(meldsettings.style_scheme)
@@ -51,27 +49,12 @@ class MeldBuffer(GtkSource.Buffer):
             self.set_style_scheme(meldsettings.style_scheme)
 
     def do_begin_user_action(self, *args):
-        self.user_action_count += 1
         if self.undo_sequence:
             self.undo_sequence.begin_group()
 
     def do_end_user_action(self, *args):
         if self.undo_sequence:
             self.undo_sequence.end_group()
-        self.user_action_count -= 1
-
-    def do_apply_tag(self, tag, start, end):
-        # Filthy, evil, horrible hack. What we're doing here is trying to
-        # figure out if a tag apply has come from a paste action, in which
-        # case GtkTextBuffer will 'helpfully' apply the existing tags in the
-        # copied selection. There appears to be no way to override this
-        # behaviour, or to hook in to the necessary paste mechanics to just
-        # request that we only get plain text or something. We're abusing the
-        # user_action notion here, because we only apply the tags we actually
-        # want in a callback.
-        if tag.props.name == 'inline' and self.user_action_count > 0:
-            return
-        return GtkSource.Buffer.do_apply_tag(self, tag, start, end)
 
     def get_iter_at_line_or_eof(self, line):
         """Return a Gtk.TextIter at the given line, or the end of the buffer.
@@ -108,8 +91,14 @@ class MeldBufferData(GObject.GObject):
         str('file-changed'): (GObject.SignalFlags.RUN_FIRST, None, ()),
     }
 
+    encoding = GObject.Property(
+        type=GtkSource.Encoding,
+        nick="The file encoding of the linked GtkSourceFile",
+        default=None,
+    )
+
     def __init__(self):
-        GObject.GObject.__init__(self)
+        super().__init__()
         self._gfile = None
         self._label = None
         self._monitor = None
@@ -119,33 +108,38 @@ class MeldBufferData(GObject.GObject):
     def reset(self, gfile):
         same_file = gfile and self._gfile and gfile.equal(self._gfile)
         self.gfile = gfile
-        self.label = self._label if same_file else self.filename
+        if same_file:
+            self.label = self._label
+        else:
+            self.label = gfile.get_parse_name() if gfile else None
         self.loaded = False
         self.savefile = None
 
     def __del__(self):
-        self._disconnect_monitor()
+        self.disconnect_monitor()
 
     @property
     def label(self):
         # TRANSLATORS: This is the label of a new, currently-unnamed file.
-        return self._label or _(u"<unnamed>")
+        return self._label or _("<unnamed>")
 
     @label.setter
     def label(self, value):
         if not value:
             return
-        encodings = (sys.getfilesystemencoding(), 'utf8')
-        self._label = fallback_decode(value, encodings, lossy=True)
+        if not isinstance(value, str):
+            log.warning('Invalid label ignored "%r"', value)
+            return
+        self._label = value
 
-    def _connect_monitor(self):
+    def connect_monitor(self):
         if not self._gfile:
             return
         monitor = self._gfile.monitor_file(Gio.FileMonitorFlags.NONE, None)
         handler_id = monitor.connect('changed', self._handle_file_change)
         self._monitor = monitor, handler_id
 
-    def _disconnect_monitor(self):
+    def disconnect_monitor(self):
         if not self._monitor:
             return
         monitor, handler_id = self._monitor
@@ -165,9 +159,9 @@ class MeldBufferData(GObject.GObject):
 
     def _handle_file_change(self, monitor, f, other_file, event_type):
         mtime = self._query_mtime(f)
-        if self._disk_mtime and mtime > self._disk_mtime:
+        if self._disk_mtime and mtime and mtime > self._disk_mtime:
             self.emit('file-changed')
-        self._disk_mtime = mtime
+        self._disk_mtime = mtime or self._disk_mtime
 
     @property
     def gfile(self):
@@ -175,16 +169,15 @@ class MeldBufferData(GObject.GObject):
 
     @gfile.setter
     def gfile(self, value):
-        self._disconnect_monitor()
+        self.disconnect_monitor()
         self._gfile = value
         self._sourcefile = GtkSource.File()
         self._sourcefile.set_location(value)
+        self._sourcefile.bind_property(
+            'encoding', self, 'encoding', GObject.BindingFlags.DEFAULT)
 
-        # This is aiming to maintain existing behaviour for filename. The
-        # behaviour is however wrong and should be fixed.
-        self.filename = value.get_path().decode('utf8') if value else None
         self.update_mtime()
-        self._connect_monitor()
+        self.connect_monitor()
 
     @property
     def sourcefile(self):
@@ -192,9 +185,7 @@ class MeldBufferData(GObject.GObject):
 
     @property
     def gfiletarget(self):
-        if self.savefile:
-            return Gio.File.new_for_path(self.savefile)
-        return self.gfile
+        return self.savefile or self.gfile
 
     @property
     def is_special(self):
@@ -210,7 +201,11 @@ class MeldBufferData(GObject.GObject):
         try:
             info = self.gfiletarget.query_info(
                 Gio.FILE_ATTRIBUTE_ACCESS_CAN_WRITE, 0, None)
-        except (AttributeError, GLib.GError):
+        except GLib.GError as err:
+            if err.code == Gio.IOErrorEnum.NOT_FOUND:
+                return True
+            return False
+        except AttributeError:
             return False
         return info.get_attribute_boolean(Gio.FILE_ATTRIBUTE_ACCESS_CAN_WRITE)
 
@@ -223,7 +218,7 @@ class MeldBufferData(GObject.GObject):
         return self._mtime == self._disk_mtime
 
 
-class BufferLines(object):
+class BufferLines:
     """Gtk.TextBuffer shim with line-based access and optional filtering
 
     This class allows a Gtk.TextBuffer to be treated as a list of lines of
@@ -250,7 +245,7 @@ class BufferLines(object):
             # this will return the last line.
             start = self.buf.get_iter_at_line_or_eof(lo)
             end = self.buf.get_iter_at_line_or_eof(hi)
-            txt = text_type(self.buf.get_text(start, end, False), 'utf8')
+            txt = self.buf.get_text(start, end, False)
 
             filter_txt = self.textfilter(txt, self.buf, start, end)
             lines = filter_txt.splitlines()
@@ -271,7 +266,7 @@ class BufferLines(object):
             if hi - lo != len(lines):
                 # These codepoints are considered line breaks by Python, but
                 # not by GtkTextStore.
-                additional_breaks = set(('\x0c', '\x85', u'\u2028'))
+                additional_breaks = set(('\x0c', '\x85', '\u2028'))
                 i = 0
                 while i < len(ends):
                     line, end = lines[i], ends[i]
@@ -298,14 +293,13 @@ class BufferLines(object):
             if not line_end.ends_line():
                 line_end.forward_to_line_end()
             txt = self.buf.get_text(line_start, line_end, False)
-            txt_filtered = self.textfilter(txt, self.buf, line_start, line_end)
-            return text_type(txt_filtered, 'utf8')
+            return self.textfilter(txt, self.buf, line_start, line_end)
 
     def __len__(self):
         return self.buf.get_line_count()
 
 
-class BufferAction(object):
+class BufferAction:
     """A helper to undo/redo text insertion/deletion into/from a text buffer"""
 
     def __init__(self, buf, offset, text):
